#include <3ds.h>
#include <stdlib.h>
#include <stdio.h>
#include "arm9payload_bin.h"
#include "brahma.h"

Result app_install(const char* path); //from FBI by Steveice10
Result app_delete(u64 titleId); 
void waitSec(int sec);
Result fileCopyOverwrite(char *src, char *des);
Result fileExist(char *file);
void backup();
void install();
void restore();
FS_Archive sdmcArchive;

s32 main (void) {
	
	gfxInitDefault(); // Initialize services
	gfxSwapBuffers(); 
	
	Result res;
	const char *path="sdmc:/boot.cia";
	sdmcArchive = (FS_Archive) {ARCHIVE_SDMC, (FS_Path) {PATH_EMPTY, 1, (u8*) ""}};
	char *dontcare;
	
	consoleInit(GFX_TOP,NULL);
	printf("miniCIA - zoogie\n\n");
	
	res=khaxInit();
	printf("khax %08X\n",(int)res);
	res=amInit();
	printf("am %08X\n",(int)res);
	
	res=AM_GetTitleProductCode(MEDIATYPE_SD, 0x0004000000193000LL, dontcare); // check if FBI is already installed
	if(res){
		if( !fileExist("sdmc:/boot.3dsx.bku") ){
			printf("Backing up files...\n");
			backup();
			printf("Installing files...\n");
			install();
		}
		
		res=app_install(path);
		
		if(res){                    //this means cia install failed so we should reboot with sig patches
			printf("We need sig patches first, so ...\n");
			printf("Executing miniPasta in 3 seconds!\n\n");
			printf("(*IF* you see this message more than once,\nhold any button to abort.)");
			waitSec(3);
			minipasta();
		}
	}
	
	if(res==0){                       //success
			printf("\nFBI.cia is installed. Restoring old files...\n\n");
			restore();
			printf("\nYou may turn off the 3ds now.\n");
	}

	while(aptMainLoop()) {
        gspWaitForVBlank();
        hidScanInput();
        
        if(hidKeysDown() & KEY_START)
            break;
		if( (hidKeysDown() &  KEY_UP) && (hidKeysDown() &  KEY_Y) ) {
			res=app_delete(0x0004000000193000LL);
			printf("FBI deleted.\n%08X\n",(int)res);
			waitSec(1);
		}
            
        gfxFlushBuffers();
        gfxSwapBuffers();
    }
	
	gfxExit();
	return 0; // Return to hbmenu
}

void minipasta(){
	brahma_init();
    load_arm9_payload_from_mem (arm9payload_bin, arm9payload_bin_size);
	firm_reboot();	
}

Result app_install(const char* path) {
	
	u8 media=MEDIATYPE_SD;  //sd
	u32 bufSize = 1024 * 1024; // 1MB
	void* buf = malloc(bufSize);
	u64 pos = 0;
	u32 bytesRead;
	Result res;
	
	AM_InitializeExternalTitleDatabase(false); //dont overwrite if db already exists
	
	FSUSER_OpenArchive(&sdmcArchive);
	Handle fileHandle;
	u64 size;

	res=FSUSER_OpenFile(&fileHandle, sdmcArchive, fsMakePath(PATH_ASCII, path + 5), FS_OPEN_READ, 0);
    if(res) goto exit;

	FSFILE_GetSize(fileHandle, &size);

	Handle ciaHandle;
	res=AM_StartCiaInstall(media, &ciaHandle);
	if(res) goto exit;
	
	FSFILE_SetSize(ciaHandle, size);

	for(pos=0; pos<size; pos+=bufSize){
		FSFILE_Read(fileHandle, &bytesRead, pos, buf, bufSize);
		FSFILE_Write(ciaHandle, NULL, pos, buf, bytesRead, FS_WRITE_FLUSH);
		printf("\rProgress %d/%d MB",(int)(pos/bufSize+1),(int)(size/bufSize+1));
	}
	printf("\n");

	res=AM_FinishCiaInstall(media, &ciaHandle);

	exit:
	free(buf);
	FSFILE_Close(fileHandle);
	FSUSER_CloseArchive(&sdmcArchive);
	return res;
}

Result app_delete(u64 titleId) {
	return AM_DeleteAppTitle(MEDIATYPE_SD, titleId);
}

void waitSec(int sec){
	for(int i=0;i<sec*60;i++){
		gfxFlushBuffers();
		gfxSwapBuffers();
		gspWaitForVBlank();
	}
}

Result fileCopyOverwrite(char *src, char *des){
	
	if(!fileExist(src))return 1;
	
	FSUSER_OpenArchive(&sdmcArchive);
	Handle srcHandle;
	Handle desHandle;
	u32 bufSize = 2 * 1024 * 1024; // 2MB
	u32 bytesRead=0;
	void* buf = malloc(bufSize);
	u64 size=0;
	u64 pos=0;
	Result res=0;
	
	res=FSUSER_DeleteFile(sdmcArchive, fsMakePath(PATH_ASCII, des + 5));
	res=FSUSER_OpenFile(&srcHandle, sdmcArchive, fsMakePath(PATH_ASCII, src + 5), FS_OPEN_READ, 0);
	res=FSUSER_OpenFile(&desHandle, sdmcArchive, fsMakePath(PATH_ASCII, des + 5), FS_OPEN_WRITE | FS_OPEN_CREATE, 0);
	
	FSFILE_Read(srcHandle, &bytesRead, pos, buf, bufSize);    //we only read 2MB as a limit
    FSFILE_Write(desHandle, NULL, pos, buf, bytesRead, FS_WRITE_FLUSH);
	
	free(buf);
	FSUSER_CloseArchive(&sdmcArchive);
	FSFILE_Close(srcHandle);
	FSFILE_Close(desHandle);
	return res;
}

Result fileExist(char *file){
	Result res=0;
	FILE *f=fopen(file,"r");
	if(f){
		res=1;
	    fclose(f);
	}
	return res;
}

void backup(){
	fileCopyOverwrite("sdmc:/boot.3dsx","sdmc:/boot.3dsx.bku");
	fileCopyOverwrite("sdmc:/menuhax_padcfg.bin","sdmc:/menuhax_padcfg.bin.bku");
}

void install(){
	fileCopyOverwrite("sdmc:/3ds/miniCIA/miniCIA.3dsx","sdmc:/boot.3dsx");
	fileCopyOverwrite("sdmc:/3ds/miniCIA/boot.cia","sdmc:/boot.cia");
	fileCopyOverwrite("sdmc:/3ds/miniCIA/menuhax_padcfg.bin","sdmc:/menuhax_padcfg.bin");
}

void restore(){
	fileCopyOverwrite("sdmc:/boot.3dsx.bku","sdmc:/boot.3dsx");
	fileCopyOverwrite("sdmc:/menuhax_padcfg.bin.bku","sdmc:/menuhax_padcfg.bin");
	FSUSER_DeleteFile(sdmcArchive, fsMakePath(PATH_ASCII, "sdmc:/boot.3dsx.bku" + 5));
	FSUSER_DeleteFile(sdmcArchive, fsMakePath(PATH_ASCII, "sdmc:/menuhax_padcfg.bin.bku" + 5));
	FSUSER_DeleteFile(sdmcArchive, fsMakePath(PATH_ASCII, "sdmc:/boot.cia" + 5));
}